/*
 * Copyright 2020-2021 Moritz Ruth and Uranos contributors
 * Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file
 */

package space.uranos.mdsp.generator

import com.google.common.base.CaseFormat
import com.jsoniter.JsonIterator
import com.squareup.kotlinpoet.*
import com.squareup.kotlinpoet.ParameterizedTypeName.Companion.parameterizedBy
import space.uranos.mdsp.JsonAny
import java.io.File

class EntitiesGenerator(
    private val workingDir: File,
    private val outputDir: File,
    private val sourcesDir: File
) {
    fun generate() {
        val entitiesJson = workingDir.resolve("entities.json").readText()
        val types = JsonIterator.deserialize(entitiesJson).asList()

        generateEntityTypes(types)
        generateEntityStubs(types)
        generateEntityTypeList(types)
    }

    private fun generateEntityTypes(types: List<JsonAny>) {
        for (entity in types) {
            val name =
                CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, entity.get("name").toString())!! + "EntityType"

            val type = TypeSpec.classBuilder(name)
                .addModifiers(KModifier.ABSTRACT)
                .primaryConstructor(FunSpec.constructorBuilder().addModifiers(KModifier.INTERNAL).build())
                .addSuperinterface(ENTITY_TYPE_TYPE)
                .addProperty(
                    PropertySpec
                        .builder("numericID", Int::class, KModifier.OVERRIDE)
                        .initializer("%L", entity.get("id").toInt())
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("id", String::class, KModifier.OVERRIDE)
                        .initializer("%S", "minecraft:" + entity.get("name").toString())
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("width", Float::class, KModifier.OVERRIDE)
                        .initializer("%Lf", entity.get("width").toFloat())
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("height", Float::class, KModifier.OVERRIDE)
                        .initializer("%Lf", entity.get("height").toFloat())
                        .build()
                )
                .build()

            FileSpec.builder(ENTITY_PACKAGE, name)
                .addType(type)
                .build()
                .writeTo(outputDir)
        }
    }

    private fun generateEntityStubs(types: List<JsonAny>) {
        for (entity in types) {
            val name =
                CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, entity.get("name").toString())!! + "Entity"

            val filePathRelativeToSourceRoot = "./${ENTITY_PACKAGE.replace(".", "/")}/$name.kt"
            if (sourcesDir.resolve(filePathRelativeToSourceRoot).exists()) continue

            val type = TypeSpec.classBuilder(name)
                .superclass(ENTITY_TYPE)
                .addProperty(
                    PropertySpec.builder("type", ENTITY_TYPE_TYPE, KModifier.OVERRIDE, KModifier.FINAL)
                        .initializer("Type")
                        .build()
                )
                .addType(
                    TypeSpec.companionObjectBuilder("Type")
                        .superclass(ClassName(ENTITY_PACKAGE, name + "Type"))
                        .build()
                )
                .build()

            FileSpec.builder(ENTITY_PACKAGE, name)
                .addType(type)
                .build()
                .writeTo(outputDir)
        }
    }

    private fun generateEntityTypeList(entities: List<JsonAny>) {
        val names = entities
            .map { it.get("name").toString() }
            .map { CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, it) + "Entity" }

        val property = PropertySpec.builder(
            "ENTITY_TYPES",
            List::class.asTypeName().parameterizedBy(ENTITY_TYPE_TYPE)
        )
            .initializer("listOf(\n${names.joinToString(",\n")}\n)")
            .build()

        FileSpec.builder(ENTITY_PACKAGE, "EntityTypes")
            .addProperty(property)
            .build()
            .writeTo(outputDir)
    }
}
