/*
 * Copyright 2020-2021 Moritz Ruth and Uranos contributors
 * Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file
 */

package space.uranos

import space.uranos.world.VoxelLocation
import space.uranos.world.World
import java.lang.IllegalArgumentException
import kotlin.math.roundToInt

/**
 * A combination of x, y and z coordinates and an orientation (yaw and pitch).
 *
 * @param yaw The yaw rotation in degrees. Must be in `[0; 360[`.
 * @param pitch The pitch rotation as a value between -90 (up) and 90 (down).
 */
data class Position(val x: Double, val y: Double, val z: Double, val yaw: Float, val pitch: Float) {
    init {
        if (yaw >= 360) throw IllegalArgumentException("yaw must be lower than 360")
        if (yaw < 0) throw IllegalArgumentException("yaw must not be negative")
    }

    /**
     * Converts this Position to a VoxelLocation by converting [x], [y] and [z] to integers using [Double.toInt],
     * in contrast to [roundToBlock] which uses [Double.roundToInt].
     */
    fun toVoxelLocation(): VoxelLocation = VoxelLocation(x.toInt(), y.toInt().toUByte(), z.toInt())

    /**
     * Converts this Position to a VoxelLocation by converting [x], [y] and [z] to integers using [Double.roundToInt],
     * in contrast to [toVoxelLocation] which uses [Double.toInt].
     */
    fun roundToBlock(): VoxelLocation = VoxelLocation(x.roundToInt(), y.roundToInt().toUByte(), z.roundToInt())

    fun toLocation(): Location = Location(x, y, z)
    fun toVector() = Vector(x, y, z)
    infix fun inside(world: World): Pair<World, Position> = world to this

    val yawIn256Steps get() = ((yaw / 360) * 256).toInt().toUByte()
    val pitchIn256Steps get() = ((yaw / 360) * 256).toInt().toUByte()

    operator fun get(part: CoordinatePart): Double = when (part) {
        CoordinatePart.X -> x
        CoordinatePart.Y -> y
        CoordinatePart.Z -> z
    }

    companion object {
        val ZERO = Position(0.0, 0.0, 0.0, 0f, 0f)
    }
}
