/*
 * Copyright 2020-2021 Moritz Ruth and Uranos contributors
 * Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file
 */

package space.uranos.world

import kotlinx.coroutines.cancel
import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import space.uranos.Uranos
import space.uranos.entity.Entity
import space.uranos.util.newSingleThreadDispatcher
import space.uranos.util.supervisorChild
import space.uranos.util.untilPossiblyNegative
import java.util.*
import kotlin.coroutines.CoroutineContext

/**
 * A Minecraft world.
 */
abstract class World(val uuid: UUID) {
    private val identifier = "World($uuid)"
    private val threadExecutor = newSingleThreadDispatcher(identifier)

    /**
     * [CoroutineContext] confined to the world thread.
     *
     * Is cancelled when the world is unloaded.
     */
    val coroutineContext: CoroutineContext = Uranos.coroutineContext.supervisorChild(identifier) + threadExecutor

    abstract val dimension: Dimension
    abstract val loadedChunks: Map<Chunk.Key, Chunk>
    abstract val isFlat: Boolean

    /**
     * This can be any value.
     *
     * Only the first 8 bytes of the SHA-256 hash of this value are sent to the client
     * (and it seems not to use it in any way).
     * A random value will be used if this is null.
     */
    open val seed: Long? = null

    /**
     * All entities in this world. Use [spawnEntity] for spawning new ones.
     */
    abstract val entities: Collection<Entity>

    abstract fun getChunk(key: Chunk.Key): Chunk

    /**
     * Returns the chunk containing the voxel at [location].
     */
    fun getChunk(location: VoxelLocation): Chunk = getChunk(Chunk.Key.from(location))

    /**
     * Returns the voxel at [location].
     */
    fun getVoxel(location: VoxelLocation): Voxel = getChunk(location).getVoxel(location)

    /**
     * Returns all voxels in the cube with the corner points [cornerA] and [cornerB].
     *
     * @param hollow Whether the cube is hollow
     */
    @OptIn(ExperimentalStdlibApi::class)
    fun getVoxelsInCube(cornerA: VoxelLocation, cornerB: VoxelLocation, hollow: Boolean = false): List<Voxel> {
        fun getList(a: Int, b: Int) =
            if (hollow) listOf(a, b).distinct()
            else (a untilPossiblyNegative b).toList()

        return buildList {
            for(x in getList(cornerA.x, cornerB.x)) {
                for(y in getList(cornerA.y.toInt(), cornerB.y.toInt())) {
                    for(z in getList(cornerA.z, cornerB.z)) {
                        add(getVoxel(VoxelLocation(x, y.toUByte(), z)))
                    }
                }
            }
        }
    }

    /**
     * Returns all voxels in a sphere with the specified [center] and [radius].
     */
    fun getVoxelsInSphere(
        center: VoxelLocation,
        radius: Int,
        hollow: Boolean
    ): List<Voxel> {
        // Implementation example: https://git.io/JLysq
        TODO()
    }

    /**
     * Spawns [entity][Entity] in this world.
     */
    abstract fun spawnEntity(entity: Entity)

    suspend fun destroy() = withContext(coroutineContext) {
        // TODO: Move or kick players
        coroutineContext.cancel()

        coroutineScope {
            loadedChunks.values.forEach { launch { (it as? Chunk.Unloadable)?.unload() } }
        }

        threadExecutor.close()
    }
}
