/*
 * Copyright 2020-2021 Moritz Ruth and Uranos contributors
 * Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file
 */

package space.uranos

import com.sksamuel.hoplite.ConfigFilePropertySource
import com.sksamuel.hoplite.ConfigLoader
import com.sksamuel.hoplite.ConfigSource
import kotlinx.coroutines.CoroutineName
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.asCoroutineDispatcher
import kotlinx.coroutines.runBlocking
import space.uranos.command.Command
import space.uranos.config.UranosConfig
import space.uranos.event.UranosEventBus
import space.uranos.event.UranosEventHandlerPositionManager
import space.uranos.logging.Logger
import space.uranos.logging.UranosLoggingOutputProvider
import space.uranos.net.UranosSocketServer
import space.uranos.plugin.UranosPluginManager
import space.uranos.recipe.Recipe
import space.uranos.server.Server
import space.uranos.util.EncryptionUtils
import space.uranos.world.BiomeRegistry
import space.uranos.world.Dimension
import java.io.File
import java.security.KeyPair
import java.util.concurrent.Executors
import kotlin.coroutines.CoroutineContext
import kotlin.system.exitProcess

// TODO: Consider using DI because this improves testability

class UranosServer internal constructor() : Server {
    val logger = Logger("Server")
    private val socketServer = UranosSocketServer(this)

    val keyPair: KeyPair =
        try {
            EncryptionUtils.generateKeyPair()
        } catch (e: Exception) {
            failInitialization(e)
        }

    val x509EncodedPublicKey: ByteArray = EncryptionUtils.generateX509Key(keyPair.public).encoded

    override val coroutineContext: CoroutineContext =
        CoroutineName("Server") + SupervisorJob() +
                Executors.newSingleThreadExecutor { r -> Thread(r, "server") }.asCoroutineDispatcher()

    override val sessions by socketServer::sessions
    override val players get() = sessions.mapNotNull { it.player }

    override val pluginManager = UranosPluginManager(this)
    override val serverDirectory: File =
        File(UranosServer::class.java.protectionDomain.codeSource.location.toURI()).let {
            it.mkdirs()
            if (VERSION == "development") it.resolve("../../../../../serverData").normalize()
            else it
        }

    override val dimensionRegistry = Registry<Dimension>()
    override val recipeRegistry = Registry<Recipe>()
    override val commandRegistry = Registry<Command>()
    override val biomeRegistry = BiomeRegistry()

    override val loggingOutputProvider = UranosLoggingOutputProvider
    override val scheduler = UranosScheduler()

    val config = ConfigLoader.Builder()
        .addPropertySource(
            ConfigFilePropertySource(
                ConfigSource.FileSource(serverDirectory.resolve("./config.yml").normalize()),
                optional = true
            )
        )
        .addPropertySource(
            ConfigFilePropertySource(
                ConfigSource.FileSource(File(javaClass.classLoader.getResource("default-config.yml")!!.toURI()))
            )
        )
        .build().loadConfigOrThrow<UranosConfig>()

    override val minimumLogLevel = config.minLogLevel
    override val developmentMode: Boolean = config.developmentMode

    override val eventBus = UranosEventBus(developmentMode)
    override val eventHandlerPositions = UranosEventHandlerPositionManager()

    private fun failInitialization(t: Throwable): Nothing {
        logger.error("Server initialization failed:", t)
        exitProcess(1)
    }

    fun start() {
        logger info "Starting UranosServer ($VERSION_WITH_V)"
        logger trace "Configuration: $config"

        pluginManager.loadAll()
        pluginManager.enableAll()

        socketServer.bind()
        logger info "Listening on ${config.host}:${config.port}"

        scheduler.startTicking()
    }

    override fun shutdown() {
        runBlocking {
            scheduler.shutdown()
        }
    }

    companion object {
        val VERSION = UranosServer::class.java.`package`.implementationVersion ?: "development"
        val VERSION_WITH_V = if (VERSION == "development") VERSION else "v$VERSION"

        @JvmStatic
        fun main(args: Array<String>) {
            UranosServer().also { setServerInstance(it) }.start()
        }

        fun setServerInstance(instance: Server) {
            val clazz = Class.forName("space.uranos.UranosKt")
            val field = clazz.getDeclaredField("serverInstance")
            field.isAccessible = true
            field.set(null, instance)
            field.isAccessible = false
        }
    }
}
