/*
 * Copyright 2020-2021 Moritz Ruth and Uranos contributors
 * Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file
 */

package space.uranos.net

import io.netty.buffer.ByteBuf
import io.netty.util.ReferenceCountUtil
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import space.uranos.UranosServer
import space.uranos.chat.ChatColor
import space.uranos.chat.ChatComponent
import space.uranos.chat.TextComponent
import space.uranos.event.*
import space.uranos.logging.Logger
import space.uranos.net.MinecraftProtocolDataTypes.readVarInt
import space.uranos.net.event.PacketReceivedEvent
import space.uranos.net.event.PacketSendEvent
import space.uranos.net.packet.IncomingPacket
import space.uranos.net.packet.OutgoingPacket
import space.uranos.net.packet.Protocol
import space.uranos.net.packet.handshaking.HandshakingProtocol
import space.uranos.net.packet.login.LoginProtocol
import space.uranos.net.packet.play.OutgoingKeepAlivePacket
import space.uranos.net.packet.play.OutgoingPluginMessagePacket
import space.uranos.net.packet.play.PlayProtocol
import space.uranos.net.packet.status.StatusProtocol
import space.uranos.server.event.SessionInitializedEvent
import space.uranos.util.awaitSuspending
import space.uranos.util.supervisorChild
import java.net.InetAddress
import java.net.InetSocketAddress
import javax.crypto.SecretKey
import kotlin.coroutines.CoroutineContext
import kotlin.properties.Delegates

class UranosSession(private val channel: io.netty.channel.Channel, val server: UranosServer) : Session() {
    override val address: InetAddress = (channel.remoteAddress() as InetSocketAddress).address

    private val identifier = "UranosSession(${address.hostAddress})"
    val logger = Logger(identifier)

    override val coroutineContext: CoroutineContext = server.coroutineContext.supervisorChild(identifier)
    val scope = CoroutineScope(coroutineContext)

    override var brand: String? = null
    override var ping: Int = -1

    override var state: State = State.WaitingForHandshake

    override val currentProtocol: Protocol?
        get() = when (state) {
            is State.WaitingForHandshake -> HandshakingProtocol

            is State.WaitingForStatusRequest -> StatusProtocol

            is State.WaitingForLoginStart,
            is State.WaitingForEncryptionVerification,
            is State.Encrypted -> LoginProtocol

            is State.LoginSucceeded,
            is State.WaitingForClientSettings,
            is State.Joining,
            is State.Playing -> PlayProtocol

            is State.Disconnected -> null
        }

    override suspend fun disconnect(reason: TextComponent, internalReason: String?) {
        stopProcessingIncomingPackets()

        if (currentProtocol == LoginProtocol || currentProtocol == PlayProtocol) {
            logger info "Disconnected" + (internalReason?.let { ". Internal reason: $it" } ?: "")

            val finalReason = if (server.config.developmentMode && internalReason != null) {
                val additional = listOf<ChatComponent>(
                    TextComponent(
                        "\n\nLogged reason (only shown in development mode): \n",
                        color = ChatColor.GRAY
                    ),
                    TextComponent of internalReason
                )

                reason.copy(extra = reason.extra + additional)
            } else reason

            when (currentProtocol) {
                LoginProtocol -> send(space.uranos.net.packet.login.DisconnectPacket(finalReason))
                PlayProtocol -> send(space.uranos.net.packet.play.DisconnectPacket(finalReason))
            }
        } else logger trace "Disconnected"

        onDisconnect(true)
        channel.close()
    }

    private var packetsChannel: Channel<IncomingPacket>? = null
    private val packetDataChannel: Channel<ByteBuf> = Channel(Channel.UNLIMITED)

    fun launchPacketsChannelConsumer() {
        val channel = Channel<IncomingPacket>(server.config.packetsBufferSize)
        packetsChannel = channel

        scope.launch {
            for (packet in channel) {
                handlePacket(packet)
            }
        }
    }

    private fun launchPacketDataChannelConsumer() {
        scope.launch {
            for (data in packetDataChannel) {
                try {
                    handlePacketData(data)
                } catch (e: Exception) {
                    failAndDisconnectBecauseOfClient(e)
                }
            }
        }
    }

    private fun stopProcessingIncomingPackets() {
        packetDataChannel.close()
        packetsChannel?.close()
    }

    private suspend fun handlePacket(packet: IncomingPacket) {
        logger.trace { "Packet received: $packet" }

        server.eventBus.emit(PacketReceivedEvent(this@UranosSession, packet)).ifNotCancelled {
            SessionPacketReceivedEventHandler.handle(this@UranosSession, packet)
        }
    }

    private suspend fun handlePacketData(data: ByteBuf) {
        val packetID = data.readVarInt()
        val codec = currentProtocol!!.incomingPacketCodecsByID[packetID]

        if (codec == null) {
            val message = "Received an unknown packet (ID: $packetID)"

            if (server.config.developmentMode)
                logger warn "$message. This will cause the client to disconnect in production mode."
            else failAndDisconnectBecauseOfClient(message)

            return
        }

        val packet = codec.decode(data)
        ReferenceCountUtil.release(data)

        if (packetsChannel != null) {
            // The protocol will not change anymore, so we can safely decode all pending packets with the current one
            if (!packetsChannel!!.offer(packet)) {
                logger.warn("The packet buffer is full (size: ${server.config.packetsBufferSize})")
                disconnect(internalReason = "You sent more packets than the packet buffer can handle")
            }
        } else {
            // The protocol could change because of this packet, so we must wait
            handlePacket(packet)
        }
    }

    val joinProcedure = LoginAndJoinProcedure(this)

    var lastKeepAlivePacketTimestamp by Delegates.notNull<Long>()
    lateinit var keepAliveDisconnectJob: Job

    fun scheduleKeepAlivePacket(isFirst: Boolean = false) {
        scope.launch {
            if (!isFirst) {
                val timeSinceLastPacket = (System.currentTimeMillis() - lastKeepAlivePacketTimestamp).toInt()
                delay(KEEP_ALIVE_PACKET_INTERVAL.toLong() - timeSinceLastPacket)
            }

            lastKeepAlivePacketTimestamp = System.currentTimeMillis()
            send(OutgoingKeepAlivePacket(lastKeepAlivePacketTimestamp))

            keepAliveDisconnectJob = launch {
                delay(server.config.timeout.toMillis())
                disconnect(
                    TextComponent of "Timed out",
                    "The client did not respond to the KeepAlive packet for " +
                            "${server.config.timeout.toMillis()}ms"
                )
            }
        }
    }

    fun onConnect() = scope.launch {
        logger trace "Connected"
        launchPacketDataChannelConsumer()

        state = State.WaitingForHandshake
        if (server.eventBus.emit(SessionInitializedEvent(this@UranosSession)).cancelled) channel.close()
        else server.sessions.add(this@UranosSession)
    }

    fun onDisconnect(expected: Boolean) {
        scope.launch {
            if (state == State.Disconnected) return@launch
            if (!expected && currentProtocol != HandshakingProtocol && currentProtocol != StatusProtocol)
                logger trace "The client disconnected unexpectedly"

            stopProcessingIncomingPackets()
            coroutineContext.cancel(DisconnectedCancellationException())
            state = State.Disconnected
            server.sessions.remove(this@UranosSession)
        }
    }

    fun onPacketDataReceived(packetData: ByteBuf) {
        packetDataChannel.offer(packetData)
    }

    internal fun failAndDisconnectBecauseOfClient(throwable: Throwable) =
        failAndDisconnectBecauseOfClient("${throwable::class.java.name}: ${throwable.message}", throwable)

    internal fun failAndDisconnectBecauseOfClient(message: String, throwable: Throwable? = null) {
        val start = "The client caused a connection error"

        if (server.config.ignoreClientCausedErrors) {
            logger debug "The client caused a connection error: $message"
            // Does not send the Disconnect packet
            channel.close()
            onDisconnect(true)
        } else {
            if (throwable == null) logger.error("$start: $message")
            else logger.error(start, throwable)

            scope.launch {
                // Sends the Disconnect packet
                disconnect(internalReason = message)
            }
        }
    }

    fun enableEncryptionCodec(key: SecretKey) {
        logger trace "Enabling encryption codec"
        if (channel.pipeline().get("encryption") != null) return
        channel.pipeline().addBefore("framing", "encryption", EncryptionCodec(this, key))
    }

    fun enableCompressionCodec() {
        logger trace "Enabling compression codec"
        if (channel.pipeline().get("compression") != null) return

        channel.pipeline()
            .addAfter("framing", "compression", CompressionCodec(server.config.packetCompressionThreshold))
    }

    override suspend fun send(packet: OutgoingPacket) = withContext<Unit>(coroutineContext) {
        logger.trace { "Sending packet: $packet" }
        server.eventBus.emit(PacketSendEvent(this@UranosSession, packet)).ifNotCancelled {
            try {
                channel.writeAndFlush(it.packet).awaitSuspending()
            } catch (t: Throwable) {
                if (!channel.isActive) return@withContext
                logger.error("Sending packet failed", t)
            }
        }
    }

    override suspend fun sendPluginMessage(channel: String, data: ByteBuf) {
        if (this.currentProtocol != PlayProtocol) throw IllegalStateException("The session is not using the PLAY protocol")
        send(OutgoingPluginMessagePacket(channel, data))
        data.release()
    }

    companion object {
        const val KEEP_ALIVE_PACKET_INTERVAL = 1000
    }
}
